/*
 * Copyright (c) 2005,2015 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <neobotix/socketcan.h>
#include <neobotix/neo.h>
#include <neobotix/canIO.h>

#include "canIOCmd.h"
#include "canIds.h"

/*----------------------------------------------------------------------*/
/*
 * Initialize communication with the I/O board
 */
CAN_IO_STR *
canIOInit(int dev)
{
	CAN_IO_STR *io;

	io = (CAN_IO_STR *)malloc(sizeof(CAN_IO_STR));
	if (io == NULL) {
		return NULL;
	}
	io->mutex = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
	if (io->mutex == NULL) {
		free(io);
		return NULL;
	}
	pthread_mutex_init(io->mutex, NULL);
	io->dev = dev;
	io->recMsgId = IOBOARD_RX_ID;
	return io;
}

/*----------------------------------------------------------------------*/
/*
 * Ends communication with the I/O board 
 */
void 
canIOEnd(CAN_IO_STR *io)
{
	pthread_mutex_destroy(io->mutex);
	free(io->mutex);
	free(io);
}

/*----------------------------------------------------------------------*/
/*
 * Send the 'connect' message to the I/O board
 * (needed before any other communication can take place)
 */
void
canIORequestConnect(NEO_STR *neo)
{
	CAN_IO_STR *io = neo->io;
	CAN_MSG m;

	canMsgSet(&m, IOBOARD_TX_ID, 
	    CMD_IOBOARD_CONNECT, 0, 0, 0, 0, 0, 0, 0);
	socketcanTransmitMsg(io->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Request status 
 */
void
canIORequestStatus(NEO_STR *neo)
{
	CAN_IO_STR *io = neo->io;
	CAN_MSG m;
	
	canMsgSet(&m, IOBOARD_TX_ID,
	    CMD_IOBOARD_GETSTATUS, 0, 0, 0, 0, 0, 0, 0);
	socketcanTransmitMsg(io->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Request battery level
 */
void
canIORequestVBatt(NEO_STR *neo)
{
	CAN_IO_STR *io = neo->io;
	CAN_MSG m;

	canMsgSet(&m, IOBOARD_TX_ID,
	    CMD_IOBOARD_GETVBATT, 0, 0, 0, 0, 0, 0, 0);
	socketcanTransmitMsg(io->dev, &m);
}
/*----------------------------------------------------------------------*/
/*
 * Request analog inputs
 */
void
canIORequestAnalogIn(NEO_STR *neo)
{
	CAN_IO_STR *io = neo->io;
	CAN_MSG m;

	canMsgSet(&m, IOBOARD_TX_ID,
	    CMD_IOBOARD_GETANALOGIN, 16, 0, 0, 0, 0, 0, 0);
	socketcanTransmitMsg(io->dev, &m);
}
/*----------------------------------------------------------------------*/
/*
 * Request digital inputs
 */
void
canIORequestDigiIn(NEO_STR *neo)
{
	CAN_IO_STR *io = neo->io;
	CAN_MSG m;

	canMsgSet(&m, IOBOARD_TX_ID,
	    CMD_IOBOARD_GETDIGIN, 0, 0, 0, 0, 0, 0, 0);
	socketcanTransmitMsg(io->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Decode CAN messages sent by the I/O board
 */
bool
canIODecode(CAN_IO_STR *io, CAN_MSG *m)
{
	int cmd = canGetCmd(m);

	if (m->id != io->recMsgId) 
		return false;
	pthread_mutex_lock(io->mutex);
	switch (cmd) {
	case CMD_IOBOARD_CONNECT:
		if (m->data[0] == 1 && m->data[1] == 2
		    && m->data[2] == 3 && m->data[3] == 4
		    && m->data[4] == 5 && m->data[5] == 6
		    && m->data[6] == 7)
			io->connected = true;
		else 
			io->connected = false;
		break;
	case CMD_IOBOARD_GETSTATUS:
		io->VBattx10 = m->data[2] << 8 | m->data[3];
		break;
	case CMD_IOBOARD_GETVBATT:
		io->VBattx10 = m->data[0] << 8 | m->data[1];
		break;
	case CMD_IOBOARD_GETANALOGIN:
		if (m->data[5] == 0) {
			io->analogIn[0] = m->data[0] 
				| ((m->data[4] & 0x03) << 8);
			io->analogIn[1] = m->data[1] 
				| ((m->data[4] & 0x0C) << 6);
			io->analogIn[2] = m->data[2] 
				| ((m->data[4] & 0x30) << 4);
			io->analogIn[3] = m->data[3] 
				| ((m->data[4] & 0xC0) << 2);
		} else if (m->data[5] == 1) {
			io->analogIn[4] = m->data[0] 
				| ((m->data[4] & 0x03) << 8);
			io->analogIn[5] = m->data[1] 
				| ((m->data[4] & 0x0C) << 6);
			io->analogIn[6] = m->data[2] 
				| ((m->data[4] & 0x30) << 4);
			io->analogIn[7] = m->data[3] 
				| ((m->data[4] & 0xC0) << 2);
		}
		break;
	case CMD_IOBOARD_GETDIGIN:
		io->digiIn = (m->data[0] << 8) | m->data[1];
		break;
	default:
		break;
	}
	pthread_mutex_unlock(io->mutex);
	return true;
}

/*----------------------------------------------------------------------*/
/* 
 * Returns current battery level
 */
int
canIOGetVBatt(NEO_STR *neo)
{
	CAN_IO_STR *io = neo->io;
	int v;

	pthread_mutex_lock(io->mutex);
	v = io->VBattx10;
	pthread_mutex_unlock(io->mutex);
	return v;
}
/*----------------------------------------------------------------------*/
/* 
 * Returns digital inputs
 */
unsigned short
canIOGetDigiIn(NEO_STR *neo)
{
	CAN_IO_STR *io = neo->io;
	unsigned short v;

	pthread_mutex_lock(io->mutex);
	v = io->digiIn;
	pthread_mutex_unlock(io->mutex);
	return v;
}
/*----------------------------------------------------------------------*/
/* 
 * Returns analog inputs
 */
void
canIOGetAnalogIn(NEO_STR *neo, unsigned short *ana)
{
	CAN_IO_STR *io = neo->io;

	pthread_mutex_lock(io->mutex);
	memcpy(ana, io->analogIn, 8*sizeof(unsigned short));
	pthread_mutex_unlock(io->mutex);
}
/*----------------------------------------------------------------------*/
/*
 * Displays I/O board information on stdout
 */
void
canIOShow(NEO_STR *neo)
{
	CAN_IO_STR *io = neo->io;
	unsigned short ana[8],i;

	printf("connected: %s\n", io->connected ? "TRUE" : "FALSE");
	printf("vBatt    : %d\n", canIOGetVBatt(neo)); 
	printf("digitalin: 0x%04x\n", canIOGetDigiIn(neo));
	canIOGetAnalogIn(neo, ana);
	printf("analogIn :");
	for (i = 0; i<8; i++) {
		printf(" %u", ana[i]);
	}
	printf("\n");
}
