/*
 * Copyright (c) 2005,2015 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include <neobotix/socketcan.h>
#include <neobotix/neo.h>
#include <neobotix/canGyro.h>
#include <neobotix/canIO.h>
#include <neobotix/canUS.h>
#include <neobotix/canMotors.h>

static void* neoReadTask(void *);

/*----------------------------------------------------------------------*/
/*
 * Init communication with the neobotix platform
 */
NEO_STR *
neoInit(const char *device)
{
	NEO_STR *n;
	CAN_MSG m;

	n = (NEO_STR *)malloc(sizeof(NEO_STR));
	if (n == NULL)
		return NULL;

	n->can = socketcanInit(device);
	if (n->can == -1) 
		return NULL;

	n->gyro = canGyroInit(n->can);
	n->io = canIOInit(n->can);
	n->us = canUSInit(n->can);
	n->motors = canMotorsInit(n->can);
	n->dying = false;

	if (pthread_create(&n->readTask, NULL, neoReadTask, n) != 0) {
		if (n->gyro) 
			canGyroEnd(n->gyro);
		socketcanEnd(n->can);
		return NULL;
	}
	/* Send init message */
	canMsgSet(&m, 0, 1, 0, 0, 0, 0, 0, 0, 0);
	m.len = 2;
	socketcanTransmitMsg(n->can, &m);
	return n;
}
/*----------------------------------------------------------------------*/
/*
 * End communication with the neobotix platform
 */
void
neoEnd(NEO_STR *neo)
{
	neo->dying = true;
	if (neo->gyro != NULL)
		canGyroEnd(neo->gyro);
	if (neo->io != NULL)
		canIOEnd(neo->io);
	if (neo->us != NULL)
		canUSEnd(neo->us);
	if (neo->motors != NULL)
		canMotorsEnd(neo->motors);
	socketcanEnd(neo->can);
	free(neo);
}

/*----------------------------------------------------------------------*/
/*
 * Body of the neoRead task, that reads and decodes CAN messages sent 
 * by the various components of the platform 
 */
static void*
neoReadTask(void *arg)
{
	NEO_STR *neo = (NEO_STR *)arg;
	CAN_MSG m;
	int n;
	bool decoded = false;

	do {
		n = socketcanReceiveMsgWait(neo->can, &m, -1);
		if (neo->dying == true) 
			break;
		decoded = false;
		if (neo->gyro != NULL) {
			decoded |= canGyroDecode(neo->gyro, &m);
		}
		if (neo->io != NULL) {
			decoded |= canIODecode(neo->io, &m);
		}
		if (neo->us != NULL) {
			decoded |= canUSDecode(neo->us, &m);
		}
		if (neo->motors != NULL) {
			decoded |= canMotorsDecode(neo->motors, &m);
		}
		if (!decoded) {
			fprintf(stderr, "%s: undecoded packet\n", __func__);
			canMsgShow(stderr, &m);
		}
	} while (n != -1);
	return NULL;
}
